-----------------------------------------------------------------------------------------------
--[[ Copyright (c) chckxy (watdametz@gmail.com). All rights reserved

This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.  --]]
-----------------------------------------------------------------------------------------------
 
local FlipGroups = {}
local FlipGroupOption = {}
local FlipGroup = {}
local FlipGroupHeader = {}
local FlipItem = {}
local FlipItemHeader = {}

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------
-- FlipGroups is the module that handles group customization

function FlipGroups:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	o:Init()
	return o
end

function FlipGroups:Init()
	-- input fields
	self.flip = self.flip
	self.nMinFontSize = self.nMinFontSize
	
	-- setup our window
	self.xmlDoc = XmlDoc.CreateFromFile("Flip.xml")	
	self.xmlGroupDoc = XmlDoc.CreateFromFile("FlipGroups.xml")	
	self.wndMain = Apollo.LoadForm(self.xmlDoc, "FlipForm", nil, self)
    self.wndMain:Show(false, true)

	-- the left and right sides
	self.wndLeft = self.wndMain:FindChild("Left")
	self.wndRight = self.wndMain:FindChild("Right")
	
	-- the tabs
	self.wndTabs = self.wndMain:FindChild("TabContainer")
	
	self.tItemsCX = {}
	self.tItemsAH = {}

	local wnd = Apollo.LoadForm(self.xmlGroupDoc, "GroupHeader", nil, self)
	self.nHeaderHeight = wnd:GetHeight()
	wnd:Destroy()
	wnd = Apollo.LoadForm(self.xmlGroupDoc, "ItemHeader", nil, self)
	self.nItemHeight = wnd:GetHeight()
	wnd:Destroy()
	
	self.tabGroup = self:AddTab(self)
end

-----------------------------------------------------------------------------------------------
-- dealing with saving and restoring

-- ToSave : -> GroupSave
-- saves the CX and AH group
function FlipGroups:ToSave()
	return { 
		tGroupCX = self.groupCX and self.groupCX:ToSave(), 
		tGroupAH = self.groupAH and self.groupAH:ToSave() 
	}
end

-- RestoreSave : GroupSave -> Boolean
-- restores the CX and AH group
function FlipGroups:RestoreSave( tGroupsSave )
	self.groupCX = FlipGroup:new({ 
		groups = self,
		groupParent = nil,
		strName = "Commodities", 
		tSave = tGroupsSave.tGroupCX,
	})
	self.groupCX:PutItems(self.tItemsCX)

	self.groupAH = FlipGroup:new({ 
		groups = self,
		groupParent = nil, 
		strName = "Auctionables", 
		tSave = tGroupsSave.tGroupAH,
	})
	self.groupAH:PutItems(self.tItemsAH)

	return true
end


-----------------------------------------------------------------------------------------------
-- GetModules : -> [Table-of Module]
-- get the modules that hook into our groups

function FlipGroups:GetModules()
	return self.flip:GetModules()
end

-----------------------------------------------------------------------------------------------
-- relating to our group options window, which can have many tabs

-- AddTab : LuaObject -> TabWindow
-- called by a module requesting that we add a tab for it
-- TODO: guard against max tabbage
function FlipGroups:AddTab( obj )
	local tab = Apollo.LoadForm(self.xmlDoc, "Tab", self.wndTabs, self)
	tab:SetText(obj:TabName())
	tab:SetData(obj)
	self.wndTabs:ArrangeChildrenHorz(0)
	return tab
end

-- OnTab : TabWindow -> 
-- when a tab is clicked, get the module and tell it to open in our groups window
function FlipGroups:OnTab( tab )
	for nKey, tab1 in ipairs(self.wndTabs:GetChildren()) do
		if tab1 ~= tab then tab1:SetOpacity(0.7) end
	end
	tab:SetOpacity(1)
	local obj = tab:GetData()
	obj:Open(self.wndMain, self.wndLeft, self.wndRight)
end

-- OnClose : ->
-- on the close button, close the window!
function FlipGroups:OnClose()
	self.wndMain:Close()
end

-----------------------------------------------------------------------------------------------
-- relating specifically to the "Groups" tab

function FlipGroups:TabName() 
	return "Groups"
end

-- OnFlipGroupOn : ->
-- opens the group customization window
function FlipGroups:Open()
	self.headerActive = nil
	
	self:CheckGroups()
	self.wndLeft:DestroyChildren()
	self.wndRight:DestroyChildren()

	self.groupCX:Draw(self.wndLeft, self)
	self.groupAH:Draw(self.wndLeft, self)

	self.wndLeft:ArrangeChildrenVert(0)
	self.wndMain:Invoke()
end

-- the slash command should pretend like we clicked on the "Groups" tab
function FlipGroups:OnFlipGroupOn()
	self:OnTab(self.tabGroup)
end

-----------------------------------------------------------------------------------------------
-- CheckGroups : ->
-- makes sure that the CX and AH groups are initialized (in case we didn't have a save to restore)

function FlipGroups:CheckGroups()
	if not self.groupCX then
		self.groupCX = FlipGroup:new({ groups = self, groupParent = nil, strName = "Commodities", })
	end

	if not self.groupAH then
		self.groupAH = FlipGroup:new({ groups = self, groupParent = nil, strName = "Auctionables", })
	end
end

-----------------------------------------------------------------------------------------------
-- getting things

-- GetCX : -> Group
-- returns the CX group
function FlipGroups:GetCX()
	self:CheckGroups()
	return self.groupCX
end

-- GetAH : -> Group
-- gets the AH group
function FlipGroups:GetAH()
	self:CheckGroups()
	return self.groupAH
end

-- GetItem ItemID : -> Item
-- returns the item with nId if we have it
function FlipGroups:GetItem( nId )
	return self.tItemsCX[nId] or self.tItemsAH[nId]
end

-----------------------------------------------------------------------------------------------

-- AddItem : ItemID ->
-- adds a commodity to the commodity group, and an auctionable to the auctionable group
function FlipGroups:AddItem( nId )
	self:CheckGroups()
	local itemData = Item.GetDataFromId(nId)
	local group = itemData:IsAuctionable() and self.groupAH or self.groupCX
	local tItems = itemData:IsAuctionable() and self.tItemsAH or self.tItemsCX

	-- if we already have the item, then ignore this
	if tItems[nId] then return true end
	
	-- otherwise, make a new item and move it into the appropriate group
	local item = FlipItem:new({ nId = nId, })
	table.insert(tItems, nId, item)
	item:Move(group)
	return true
end

function FlipGroups:MakeItem( nId )
	return FlipItem:new({ nId = nId, groups = self, })
end

-- RemoveItem : ItemID -> 
-- remove an item by nId
function FlipGroups:RemoveItem( nId )
	local itemData = Item.GetDataFromId(nId)
	local tItems = itemData:IsAuctionable() and self.tItemsAH or self.tItemsCX
	tItems[nId] = nil
	self.wndLeft:ArrangeChildrenVert(0)
	return true
end

-----------------------------------------------------------------------------------------------

-- AddGroupWindow : String Window LuaObject -> Window
-- make a new window, but don't handle it
function FlipGroups:AddWindow( strName, wndParent, objHandler )
	local wnd = Apollo.LoadForm(self.xmlGroupDoc, strName, wndParent, objHandler)
	if strName == "ItemHeader" then
		wnd:FindChild("ItemName"):SetFont("CRB_Interface" .. self.nMinFontSize)
	end
	return wnd
end

function FlipGroups:SetMinFontSize( nSize )
	self.nMinFontSize = nSize
end

-- Redraw : ->
-- basically just redraws headerActive
function FlipGroups:Redraw()
	self.headerActive:Redraw()
end

-- RedrawItems : ->
-- basically just redraws the items in headerActive
function FlipGroups:RedrawItems()
	self.headerActive:CollapseItems()
	self.headerActive:ExpandItems()
end

-- OnDeleteActiveHeader : ->
-- when something we've done has destroyed our active group, we have to
-- do some cleanup of the left and right windows
function FlipGroups:OnDeleteActiveHeader()
	self.wndRight:DestroyChildren()
	self.headerActive:OnDestroy()
	self.wndLeft:ArrangeChildrenVert(0)
	self.headerActive = nil
end

-- OnGroupHeaderButtonDown : Button Window eType -> 
-- when we select a group in the left window, display its options in the right
-- and toggle displaying its members
function FlipGroups:OnGroupHeaderButtonDown( wndHandler )
	local header = wndHandler:GetParent():GetData()

	-- if we're clicking on the active group again, toggle displaying the members
	if header == self.headerActive then
		header:ToggleItems()

	-- otherwise, we're clicking on a different group, so display new group options
	-- and toggle its members on
	else
		if self.headerActive then
			self.headerActive:Unfocus()
			self.headerActive:CollapseItems()
		end
		self.headerActive = header
		header:Focus()
		header:ExpandSubgroups()
		header:ExpandItems()
		self:DisplayOptions(header:GetGroup())
	end
	self.wndLeft:ArrangeChildrenVert(0)
end

-- ToggleGroupOptions : Group ->
-- when a group is selected in our left window, display its options on our right window
function FlipGroups:DisplayOptions( group )
	-- make old options go byebye
	self.wndRight:DestroyChildren()

	-- auctionable have no settings yet
	if group == self.groupAH then return nil end

	-- display options
	for nKey, eType in ipairs({ "MARKET", "GROUPS", }) do
		local options = group:GetOptions(eType) 
		if options then options:Display(self.wndRight) end
	end

	-- redraw the right window
	self.wndRight:ArrangeChildrenVert(0)
end

-- GetOptions : Group -> GroupOption
-- creates new options for the group module pertaining to a specific group
function FlipGroups:GetOptions( group )
	return FlipGroupOption:new({ group = group, groups = self })
end

function FlipGroups:PullUpOptions(item, tGroups)
	self:OnTab(self.tabGroup)
	
	local fnOpenGroups 
	fnOpenGroups = function(wnd)
		if #tGroups > 0 then
			-- wnd is a window with a header as data
			for nKey, wnd in ipairs(wnd:GetChildren()) do
				local header = wnd:GetData()
				if header:GetGroup() == tGroups[#tGroups] then
					self:OnGroupHeaderButtonDown(wnd:FindChild("GroupHeaderButton"))
					table.remove(tGroups)
					fnOpenGroups(wnd:FindChild("GroupMembers"))
					break
				end
			end
		else
			for nKey, wnd in ipairs(wnd:GetChildren()) do
				local header = wnd:GetData()
				if header:GetItem() == item then
					self:OnGroupItemButton(wnd:FindChild("GroupItemButton"))
					break
				end
			end
		end
	end
	
	fnOpenGroups(self.wndLeft)
end

function FlipGroups:OnGroupItemButton( btn )
	local header = btn:GetParent():GetData()
	
	self.wndRight:DestroyChildren()
	self.headerActive:Unfocus()
	self.headerActive = header
	header:Focus()
	local wndRow = Apollo.LoadForm(self.xmlGroupDoc, "GroupOptionsRow", self.wndRight, self)
	wndRow:FindChild("Title"):SetText("Delete item:")
	wndRow:FindChild("GroupOptionButton"):SetText("Go")
end

function FlipGroups:OnGroupOptionToggle( btn )
	self:OnDeleteActiveHeader()
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------
-- GroupOptions control the group options for a specific group

function FlipGroupOption:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	o:Init()
	return o
end

function FlipGroupOption:Init()
	-- initial fields
	self.group = self.group
	self.groups = self.groups 
end

-- ToSave : -> 
-- we don't have any persistent data in our group options
function FlipGroupOption:ToSave()
	return nil
end

-- Display : Window -> 
-- displays the group options in wndParent
-- TODO: this leaks memory, so we should add a Destroy method to this object
-- that gets called when a new option tries to display
function FlipGroupOption:Display( wndParent )
	-- draw the adding subgroup box
	local wndAdd = self.groups:AddWindow("GroupOptionsRowRename", wndParent, self)
	wndAdd:FindChild("Title"):SetText("Create subgroup with name:")
	wndAdd:SetData(self.OnAddSubgroupConfirm)

	local groupParent = self.group:GetParent()
	if not groupParent then return nil end

	-- draw the rename subgroup box (can't rename the commodities/AH group)
	local wndAdd = self.groups:AddWindow("GroupOptionsRowRename", wndParent, self)
	wndAdd:FindChild("Title"):SetText("Rename this group:")
	wndAdd:SetData(self.OnRenameGroupConfirm)
	
	-- draw the exchanging group members box if we have a parent
	self.wndExchange = self.groups:AddWindow("GroupOptionsRowMembers", wndParent, self)
	local wndLeft = self.wndExchange:FindChild("Left")
	local wndRight = self.wndExchange:FindChild("Right")

	-- draw parent subgroup headers on the left
	groupParent:ForEachSubgroup(function(group)
		if group ~= self.group then group:Draw(wndLeft, self) end
	end)

	-- draw group subgroup headers on the right
	self.group:ForEachSubgroup(function(group)
		if group ~= self.group then group:Draw(wndRight, self) end
	end)

	-- then draw parent items on the left
	groupParent:ForEachItem(function(item) item:Draw(wndLeft, self) end)

	-- aand group items on the right
	self.group:ForEachItem(function(item) item:Draw(wndRight, self) end)

	-- draw the deleting box (can't delete commodities/AH group)
	local wndDel = self.groups:AddWindow("GroupOptionsRowConfirm", wndParent, self)
	wndDel:FindChild("Title"):SetText("Delete this subgroup:")
	local btn = wndDel:FindChild("GroupOptionButton")
	btn:SetData(false)
	self:OnGroupOptionToggle(btn)

	wndLeft:ArrangeChildrenVert(0)
	wndRight:ArrangeChildrenVert(0)
end

function FlipGroupOption:OnEditBoxConfirm( wndHandler )
	local fnHandler = wndHandler:GetParent():GetParent():GetData()
	fnHandler(self, wndHandler)
end

function FlipGroupOption:OnAddSubgroupConfirm( wndHandler )
	local wndRow = wndHandler:GetParent():GetParent()
	local strText = wndRow:FindChild("GroupName"):GetText()
	local subgroup = FlipGroup:new({ strName = strText, groups = self.groups, })
	subgroup:Move(self.group)
	wndRow:FindChild("GroupName"):SetText("")
	self.groups:Redraw()
end

function FlipGroupOption:OnRenameGroupConfirm( wndHandler )
	local wndRow = wndHandler:GetParent():GetParent()
	local strText = wndRow:FindChild("GroupName"):GetText()
	self.group:Rename(strText)
	wndRow:FindChild("GroupName"):SetText("")
	self.groups:Redraw()
end

-- delete the subgroup and move all the items into the parent
function FlipGroupOption:OnGroupOptionToggle( btn )
	local bSaveItems = btn:GetData()
	if bSaveItems then
		btn:SetText("Delete all members with it")
	else
		btn:SetText("Move members into parent group")
	end
	btn:SetData(not bSaveItems)
end

function FlipGroupOption:OnOptionConfirm( btn )
	local bSaveItems = btn:GetParent():FindChild("GroupOptionButton"):GetData()
	self.group:Destroy(bSaveItems)
	self.groups:OnDeleteActiveHeader()
end

-- we've clicked an item, so move it from left to right, or vice versa
function FlipGroupOption:OnGroupItemButton( btn )
	local wndItem = btn:GetParent()
	local item = wndItem:GetData():GetItem()
	local group = self.group
	local groupParent = self.group:GetParent()
	local bLeftToRight = wndItem:GetParent():GetName() == "Left"
	
	local wndLeft = self.wndExchange:FindChild("Left")
	local wndRight = self.wndExchange:FindChild("Right")
	local vVScrollLeft = wndLeft:GetVScrollPos()
	local nVScrollRight = wndRight:GetVScrollPos()

	-- remove from groupParent and add to group
	if bLeftToRight then
		item:Move(group)
		wndItem:Destroy()
		item:Draw(wndRight, self)

	-- remove from group and add to groupParent
	else 
		item:Move(groupParent)
		wndItem:Destroy()
		item:Draw(wndLeft, self)
	end
	self.groups:RedrawItems()

	wndLeft:ArrangeChildrenVert(0)
	wndRight:ArrangeChildrenVert(0)
	wndLeft:SetVScrollPos(vVScrollLeft)
	wndRight:SetVScrollPos(nVScrollRight)
end

-- we've clicked an item, so move it from left to right, or vice versa
function FlipGroupOption:OnGroupHeaderButtonDown( btn )
	local wndGroup = btn:GetParent()
	local subgroup = wndGroup:GetData():GetGroup()
	local group = self.group
	local groupParent = self.group:GetParent()
	local bLeftToRight = wndGroup:GetParent():GetName() == "Left"

	local wndLeft = self.wndExchange:FindChild("Left")
	local wndRight = self.wndExchange:FindChild("Right")
	local vVScrollLeft = wndLeft:GetVScrollPos()
	local nVScrollRight = wndRight:GetVScrollPos()

	-- remove from groupParent and add to group
	if bLeftToRight then
		subgroup:Move(group)
		wndGroup:Destroy()
		subgroup:Draw(wndRight, self)

	-- remove from group and add to groupParent
	else 
		subgroup:Move(groupParent)
		wndGroup:Destroy()
		subgroup:Draw(wndLeft, self)
	end
	self.groups:Redraw()
	
	wndLeft:ArrangeChildrenVert(0)
	wndRight:ArrangeChildrenVert(0)
	wndLeft:SetVScrollPos(vVScrollLeft)
	wndRight:SetVScrollPos(nVScrollRight)
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------
-- FlipGroupHeader is a group

function FlipGroup:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	o:Init()
	return o
end

function FlipGroup:Init()
	-- we get a parent (FlipGroupHeader or nil)
	self.groupParent = self.groupParent

	-- we have the groups instance
	self.groups = self.groups

	-- we might have a save
	self.tSave = self.tSave or {}

	-- our members, items+subgroups
	self.tItems = {}
	self.tSubgroups = {}

	-- recover group members (older save verions)
	for nKey, nId in ipairs(self.tSave.tMemberSave or {}) do
		local item = FlipItem:new({ nId = nId })
		item:Move(self)
	end

	-- recover items
	for nKey, nId in ipairs(self.tSave.tItems or {}) do
		local item = FlipItem:new({ nId = nId })
		item:Move(self)
	end

	-- recover subgroups
	for nKey, tGroupSave in ipairs(self.tSave.tSubgroups or {}) do
		local subgroup = FlipGroup:new({ tSave = tGroupSave, groups = self.groups })
		subgroup:Move(self)
	end

	-- our options
	self.options = {}
	local tOptionSave = self.tSave.tOptionSave or {}
	for eType, module in pairs(self.groups:GetModules()) do
		self.options[eType] = module:GetOptions(self, tOptionSave[eType])
	end

	-- we might have a name
	self.strName = self.strName or self.tSave.strName or ""

	self.tSave = nil
end

function FlipGroup:ToSave()
	local tItems = {}
	for nKey, member in ipairs(self.tItems) do
		table.insert(tItems, member:ToSave())
	end

	local tSubgroups = {}
	for nKey, member in ipairs(self.tSubgroups) do
		table.insert(tSubgroups, member:ToSave())
	end

	local tOptionSave = {}
	for eType, option in pairs(self.options) do
		tOptionSave[eType] = option:ToSave()
	end

	return {
		strName = self.strName,
		tItems = tItems,
		tSubgroups = tSubgroups,
		tOptionSave = tOptionSave,
	}
end

function FlipGroup:AddSubgroup(subgroup)
	table.insert(self.tSubgroups, subgroup)
	return subgroup
end

function FlipGroup:AddItem( item )
	table.insert(self.tItems, item)
	return item
end

function FlipGroup:AddWindow( strName, wndParent, objHandler)
	return self.groups:AddWindow(strName, wndParent, objHandler)
end

function FlipGroup:Draw( wndParent, objHandler )
	return FlipGroupHeader:new({ 
		objHandler = objHandler, 
		group = self,
		wndParent = wndParent,
	})
end

function FlipGroup:ForEachSubgroup( fn )
	-- have I mentioned that I hate side effects
	local tMembers = {}
	for nKey, member in ipairs(self.tSubgroups) do table.insert(tMembers, member) end
	for nKey, member in ipairs(tMembers) do fn(member) end
end

function FlipGroup:OrmapSubgroup( fn )
	-- have I mentioned that I hate side effects
	local tMembers = {}
	for nKey, member in ipairs(self.tSubgroups) do table.insert(tMembers, member) end
	for nKey, member in ipairs(tMembers) do local b = fn(member) if b then return b end end
end

function FlipGroup:ForEachItem( fn )
	-- have I mentioned that I hate side effects
	local tMembers = {}
	for nKey, member in ipairs(self.tItems) do table.insert(tMembers, member) end
	for nKey, member in ipairs(tMembers) do fn(member) end
end

function FlipGroup:OrmapItem( fn )
	-- have I mentioned that I hate side effects
	local tMembers = {}
	for nKey, member in ipairs(self.tItems) do table.insert(tMembers, member) end
	for nKey, member in ipairs(tMembers) do local b = fn(member) if b then return b end end
end

function FlipGroup:RegisterOptions( options )
	table.insert(self.options, options)
end

function FlipGroup:GetName()
	return self.strName
end

function FlipGroup:GetOptions( eType )
	return self.options[eType]
end

function FlipGroup:GetParent()
	return self.groupParent
end

function FlipGroup:Move( group )
	if self.groupParent then self.groupParent:RemoveSubgroup(self) end
	self.groupParent = group
	return group:AddSubgroup(self)
end

-- UpdateName : String ->
-- sets the name variable and displays it
function FlipGroup:Rename( strName )
	self.strName = strName or "--"
end

function FlipGroup:GetItem( nId )
	local fnGetItem = function(item)
		if item:GetItemId() == nId then return item end
	end
	local fnGetGroupItem
	fnGetGroupItem = function(subgroup)
		return subgroup:OrmapItem(fnGetItem) or
			   subgroup:OrmapSubgroup(fnGetGroupItem)
	end
	return fnGetGroupItem(self)
end

function FlipGroup:PutItems(items)
	for nKey, member in ipairs(self.tItems) do
		member:PutItems(items)
	end
	for nKey, member in ipairs(self.tSubgroups) do
		member:PutItems(items)
	end
end

function FlipGroup:RemoveItem( item )
	for nKey, member in ipairs(self.tItems) do
		if member == item then
			table.remove(self.tItems, nKey)
			return nil
		end
	end
end

function FlipGroup:RemoveSubgroup( group )
	for nKey, member in ipairs(self.tSubgroups) do
		if member == group then
			table.remove(self.tSubgroups, nKey)
			return nil
		end
	end
end

function FlipGroup:Destroy( bSaveItems )
	if bSaveItems then 
		self:ForEachItem(function(item) item:Move(self.groupParent) end)
		self:ForEachSubgroup(function(subgroup) subgroup:Move(self.groupParent) end)
	else
		self:ForEachItem(function(item) self.groups:RemoveItem(item:GetItemId()) end)
		self:ForEachSubgroup(function(subgroup) subgroup:Destroy() end)
	end
	self.groupParent:RemoveSubgroup(self)
end

function FlipGroup:PullUpOptions(item, tGroups)
	tGroups = tGroups or {}
	table.insert(tGroups, self)
	if self.groupParent then
		self.groupParent:PullUpOptions(item, tGroups)
	else
		self.groups:PullUpOptions(item, tGroups)
	end
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipGroupHeader:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	o:Init()
	return o
end

function FlipGroupHeader:Init()
	-- we have a group
	self.group = self.group
	
	-- we make window good
	self.wnd = self.group:AddWindow("GroupHeader", self.wndParent, self.objHandler)
	self.wnd:SetData(self)
	self:Unfocus()
	self.wndParent = nil
	
	-- our members
	self.wndMembers = self.wnd:FindChild("GroupMembers")

	-- our default height
	self.nHeight = self.wnd:GetHeight()

	-- setup the window
	self.wnd:FindChild("GroupName"):SetText(self.group:GetName())

	-- are our items showing?
	self.bItems = false
	self.tItemHeaders = {}
	
	-- are our subgroups showing?
	self.bSubgroups = false
	self.tSubgroupHeaders = {}
end

function FlipGroupHeader:GetGroup()
	return self.group
end

function FlipGroupHeader:GetHeight()
	return self.wnd:GetHeight()
end

function FlipGroupHeader:Redraw()
	self.wnd:FindChild("GroupName"):SetText(self.group:GetName())
	self:CollapseItems()
	self:CollapseSubgroups()
	self:ExpandSubgroups()
	self:ExpandItems()
end

function FlipGroupHeader:Unfocus()
	-- self.wnd:SetOpacity(0.8)
	self.wnd:SetSprite("BK3:UI_BK3_Holo_InsetFlyout")
end

function FlipGroupHeader:Focus()
	self.wnd:SetSprite("BK3:UI_BK3_Holo_InsetFramed_Darker")
	-- self.wnd:SetOpacity(1)
end

function FlipGroupHeader:OnDestroy()
	local headerParent = self.wnd:GetParent():GetParent():GetData()
	headerParent:CollapseSubgroups()
	headerParent:ExpandSubgroups()
end

function FlipGroupHeader:Destroy()
	self.wnd:Destroy()
end

local AddHeight
AddHeight = function(wnd, nHeight)
	if wnd:GetName() == "GroupHeader" then
		local nLeft, nTop, nRight, nBottom = wnd:GetAnchorOffsets()
		wnd:SetAnchorOffsets(nLeft, nTop, nRight, nBottom + nHeight)
		wnd:FindChild("GroupMembers"):ArrangeChildrenVert(0)
		AddHeight(wnd:GetParent():GetParent(), nHeight)
	end
end

local SubtractHeight
SubtractHeight = function(wnd, nHeight)
	if wnd:GetName() == "GroupHeader" then
		local nLeft, nTop, nRight, nBottom = wnd:GetAnchorOffsets()
		wnd:SetAnchorOffsets(nLeft, nTop, nRight, nBottom - nHeight)
		wnd:FindChild("GroupMembers"):ArrangeChildrenVert(0)
		SubtractHeight(wnd:GetParent():GetParent(), nHeight)
	end
end

	--[[local wndMembers = wndHeader:FindChild("GroupMembers")
	local bToggled = wndHeader:FindChild("GroupHeaderButton"):GetData()

	wndMembers:DestroyChildren()
	SubtractHeight(wndHeader, wndMembers:GetHeight())
	
	if bToggleOn then
		-- draw a header for each subgroup
		group:ForEachSubgroup(function(subgroup) subgroup:Draw(wndMembers, self) end)
		local nSubgroups = #wndMembers:GetChildren()

		-- draw an item for each item
		group:ForEachItem(function(item) item:Draw(wndMembers, self) end)
		local nItems = #wndMembers:GetChildren() - nSubgroups

		local nHeight = nSubgroups * self.nHeaderHeight + nItems * self.nItemHeight
		AddHeight(wndHeader, nHeight)
	end--]]

function FlipGroupHeader:ExpandSubgroups()
	if self.bSubgroups then return nil end

	self.bSubgroups = true
	local wndMembers = self.wndMembers

	-- draw an item for each item
	self.group:ForEachSubgroup(function(subgroup)
		table.insert(self.tSubgroupHeaders, subgroup:Draw(wndMembers, self.objHandler))
	end)
	local nSubgroups = #wndMembers:GetChildren()

	-- increase the height of the window and all windows above
	-- assumes all members are the same height
	if self.tSubgroupHeaders[1] then
		local nHeight = nSubgroups * self.tSubgroupHeaders[1]:GetHeight()
		AddHeight(self.wnd, nHeight)
	end
end

function FlipGroupHeader:CollapseSubgroups()
	if not self.bSubgroups then return nil end

	self.bSubgroups = false
	local wndMembers = self.wndMembers

	local nHeight = 0
	for nKey, tHeader in ipairs(self.tSubgroupHeaders) do
		nHeight = nHeight + tHeader:GetHeight()
		tHeader:Destroy()
	end
	self.tSubgroupHeaders = {}
	SubtractHeight(self.wnd, nHeight)
end

function FlipGroupHeader:ToggleItems()
	if self.bItems then self:CollapseItems() else self:ExpandItems() end
end

function FlipGroupHeader:ExpandItems()
	if self.bItems then return nil end

	self.bItems = true
	local wndMembers = self.wndMembers

	-- make room in this and every higher window for our items
	local nSubgroups = #wndMembers:GetChildren()

	-- draw an item for each item
	self.group:ForEachItem(function(item) 
		table.insert(self.tItemHeaders, item:Draw(wndMembers, self.objHandler))
	end)
	local nItems = #wndMembers:GetChildren() - nSubgroups

	-- increase the height of the window and all windows above
	-- assumes all members are the same height
	if self.tItemHeaders[1] then
		local nHeight = nItems * self.tItemHeaders[1]:GetHeight()
		AddHeight(self.wnd, nHeight)
	end
end

function FlipGroupHeader:CollapseItems()
	if not self.bItems then return nil end
	
	self.bItems = false
	local wndMembers = self.wndMembers
	
	local nHeight = 0
	for nKey, header in ipairs(self.tItemHeaders) do
		nHeight = nHeight + header:GetHeight()
		header:Destroy()
	end
	self.tItemHeaders = {}
	SubtractHeight(self.wnd, nHeight)
end

------------------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipItem:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	o:Init()
	return o
end

function FlipItem:Init()
	-- we get an id in
	self.nId = self.nId

	-- get the item data
	self.itemData = self.itemData or Item.GetDataFromId(self.nId)
	
	-- we have a parent OR we're just directly under FlipGroups
	self.groupParent = self.groupParent
	self.groups = self.groups
end

function FlipItem:Draw( wndParent, objHandler, itemData )
	return FlipItemHeader:new({ 
		item = self, 
		wndParent = wndParent, 
		objHandler = objHandler,
		itemData = itemData or self.itemData,
	})
end

function FlipItem:GetData()
	return self.itemData
end

function FlipItem:GetItemId()
	return self.nId
end

function FlipItem:GetOptions( eType )
	return self.groupParent:GetOptions(eType)
end

function FlipItem:PutItems(items)
	table.insert(items, self.nId, self)
end

function FlipItem:ToSave()
	return self.itemData:GetItemId()
end

function FlipItem:Move( group )
	if self.groupParent then self.groupParent:RemoveItem(self) end
	self.groupParent = group
	group:AddItem(self)
end

function FlipItem:Destroy()
	self.groupParent:RemoveItem(self)
end

function FlipItem:AddWindow( strName, wndParent, objHandler )
	local obj = self.groupParent or self.groups
	return obj:AddWindow(strName, wndParent, objHandler)
end

function FlipItem:PullUpOptions()
	self.groupParent:PullUpOptions(self)
end

function FlipItem:IsInGroup()
	return self.groups and self.groups:GetItem(self.nId)
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------
-- window manifestation of an item 

function FlipItemHeader:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	o:Init()
	return o
end

function FlipItemHeader:Init()
	-- we have an item
	self.item = self.item
	self.itemData = self.itemData or self.item:GetData()

	-- we make window good
	local wnd = self.item:AddWindow("ItemHeader", self.wndParent, self.objHandler)
	wnd:SetData(self)
	local wndPicture = self.item:AddWindow("ItemPictureContainer", wnd, self)
	wnd:FindChild("ItemName"):SetText(self.itemData:GetName())
	wndPicture:FindChild("ItemPicture"):SetSprite(self.itemData:GetIcon())
	wndPicture:FindChild("ItemPicture"):SetData(self.itemData)
	self.wnd = wnd

	self.nHeight = wnd:GetHeight()
	
	-- are our members showing? jeez keep it pg13
	self.bMembers = true
	
	-- not squashed, we very show
	self.bSquashed = false
end

function FlipItemHeader:SetBorderColor( strColor )
	self.wnd:SetBGColor(strColor)
end

function FlipItemHeader:GetWindow()
	return self.wnd
end

function FlipItemHeader:GetHeight()
	return self.wnd:GetHeight()
end

function FlipItemHeader:GetItem()
	return self.item
end

function FlipItemHeader:OnDestroy()
	local headerParent = self.wnd:GetParent():GetParent():GetData()
	self.wnd:Destroy()
	self.item:Destroy()
	headerParent:CollapseItems()
	headerParent:ExpandItems()
end

function FlipItemHeader:Destroy()
	self.wnd:Destroy()
end

function FlipItemHeader:Focus()
	self.wnd:SetSprite("BK3:UI_BK3_ItemQualityWhite")
	self.wnd:SetBGColor("xkcdAlgae")
	-- self.wnd:SetOpacity(1)
end

function FlipItemHeader:Unfocus()
	self.wnd:SetSprite()
	-- self.wnd:SetOpacity(0.8)
end

function FlipItemHeader:CollapseItems()
end

function FlipItemHeader:OnGenerateTooltipListIcon( wndHandler, wndControl, eType, nX, nY )
	local tCurrItem = wndHandler:GetData()
	if tCurrItem then -- moving the mouse around a lot throws errors here....
		Tooltip.GetItemTooltipForm(self, wndHandler, tCurrItem, {itemCompare = tCurrItem:GetEquippedItemForItemType()})
	end
end

function FlipItemHeader:ToggleSquash( bSquashed )
	if bSquashed ~= self.bSquashed then
		self.bSquashed = bSquashed
		if bSquashed then
			local nLeft, nTop, nRight, nBottom = self.wnd:GetAnchorOffsets()
			self.wnd:SetAnchorOffsets(nLeft, nTop, nRight, nTop)
		else
			self:ExpandMembers()
		end
	end
end

function FlipItemHeader:AddMember( objHandler )
	return self.item:AddWindow("ItemHeader", self.wnd:FindChild("ItemMembers"), self.objHandler)
end

function FlipItemHeader:ToggleMembers()
	if self.bMembers then self:CollapseMembers() else self:ExpandMembers() end
end

function FlipItemHeader:ExpandMembers()
	if self.bSquashed then return nil end 
	local nHeight = 0
	for nKey, wnd in ipairs(self.wnd:FindChild("ItemMembers"):GetChildren()) do
		nHeight = nHeight + wnd:GetHeight()
	end
	local nLeft, nTop, nRight, nBottom = self.wnd:GetAnchorOffsets()
	self.wnd:SetAnchorOffsets(nLeft, nTop, nRight, nTop + self.nHeight + nHeight)
	self.wnd:FindChild("ItemMembers"):ArrangeChildrenVert(0)
	self.bMembers = true
end

function FlipItemHeader:CollapseMembers()
	if self.bSquashed then return nil end 
	local wndMembers = self.wnd:FindChild("ItemMembers")
	local nLeft, nTop, nRight, nBottom = self.wnd:GetAnchorOffsets()
	self.wnd:SetAnchorOffsets(nLeft, nTop, nRight, nTop + self.nHeight)
	self.bMembers = false
end

function FlipItemHeader:GetSortKey()
	return self.nSortKey or 0
end

function FlipItemHeader:SetSortKey( nSortKey )
	self.nSortKey = nSortKey
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

Apollo.RegisterPackage(FlipGroups, "FlipGroups", 1, {})
